# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
#
# Babel-EDTF is free software; you can redistribute it and/or modify it under
# the terms of the MIT License; see LICENSE file for more details.

"""Module tests."""


def test_version():
    """Test version import."""
    from babel_edtf import __version__
    assert __version__



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020-2024 CERN.
#
# Babel-EDTF is free software; you can redistribute it and/or modify it under
# the terms of the MIT License; see LICENSE file for more details.

"""Module tests."""

from datetime import datetime

import pytest
from edtf import Date, DateAndTime, Interval
from edtf.parser.edtf_exceptions import EDTFParseException

from babel_edtf import edtf_to_datetime, format_edtf, parse_edtf, \
    parse_edtf_level0

separator = '\u2009–\u2009'


@pytest.mark.parametrize('edtfstr,locale,format,expected', [
    # ## Dates
    # Year (en)
    ('2020', 'en', 'short', '2020'),
    ('2020', 'en', 'medium', '2020'),
    ('2020', 'en', 'long', '2020'),
    ('2020', 'en', 'full', '2020'),
    # Year (da)
    ('2020', 'da', 'short', '2020'),
    ('2020', 'da', 'medium', '2020'),
    ('2020', 'da', 'long', '2020'),
    ('2020', 'da', 'full', '2020'),
    # Year-Month (en)
    ('2020-09', 'en', 'short', '9/2020'),
    ('2020-09', 'en', 'medium', 'Sep 2020'),
    ('2020-09', 'en', 'long', 'September 2020'),
    ('2020-09', 'en', 'full', 'September 2020'),
    # Year-Month (da)
    ('2020-09', 'da', 'short', '9.2020'),
    ('2020-09', 'da', 'medium', 'sep. 2020'),
    ('2020-09', 'da', 'long', 'september 2020'),
    ('2020-09', 'da', 'full', 'september 2020'),
    # Year-Month-Day (en)
    ('2020-09-30', 'en', 'short', '9/30/20'),
    ('2020-09-30', 'en', 'medium', 'Sep 30, 2020'),
    ('2020-09-30', 'en', 'long', 'September 30, 2020'),
    ('2020-09-30', 'en', 'full', 'Wednesday, September 30, 2020'),
    # Year-Month-Day (da)
    ('2020-09-30', 'da', 'short', '30.09.2020'),
    ('2020-09-30', 'da', 'medium', '30. sep. 2020'),
    ('2020-09-30', 'da', 'long', '30. september 2020'),
    ('2020-09-30', 'da', 'full', 'onsdag den 30. september 2020'),
    # ### Intervals (same precision)
    # Year (en)
    ('2020/2021', 'en', 'short', f'2020{separator}2021'),
    ('2020/2021', 'en', 'medium', f'2020{separator}2021'),
    ('2020/2021', 'en', 'long', f'2020{separator}2021'),
    ('2020/2021', 'en', 'full', f'2020{separator}2021'),
    # Year (da)
    ('2020/2021', 'da', 'short', '2020–2021'),
    ('2020/2021', 'da', 'medium', '2020–2021'),
    ('2020/2021', 'da', 'long', '2020–2021'),
    ('2020/2021', 'da', 'full', '2020–2021'),
    # Year-Month (constant year) (en)
    ('2020-09/2020-11', 'en', 'short', f'9/2020{separator}11/2020'),
    ('2020-09/2020-11', 'en', 'medium', f'Sep{separator}Nov 2020'),
    ('2020-09/2020-11', 'en', 'long', f'September{separator}November 2020'),
    ('2020-09/2020-11', 'en', 'full', f'September{separator}November 2020'),
    # Year-Month (different year) (en)
    ('2020-09/2021-11', 'en', 'short', f'9/2020{separator}11/2021'),
    ('2020-09/2021-11', 'en', 'medium', f'Sep 2020{separator}Nov 2021'),
    (
        '2020-09/2021-11',
        'en',
        'long',
        f'September 2020{separator}November 2021'
    ),
    (
        '2020-09/2021-11',
        'en',
        'full',
        f'September 2020{separator}November 2021'
    ),
    # Year-Month (reverse chronological order) (da)
    ('2021-11/2020-09', 'da', 'full', 'november 2021–september 2020'),
    # Year-Month-Day (en)
    ('2020-09-01/2020-11-15', 'en', 'short', f'9/1/2020{separator}11/15/2020'),
    ('2020-09-01/2020-11-15', 'en', 'medium', f'Sep 1{separator}Nov 15, 2020'),
    # Note the next two lines. For date intervals we are forced to use
    # format_skeleton, and thus these lines differ in format from other
    # long/full formats.
    ('2020-09-01/2020-11-15', 'en', 'long', f'Sep 1{separator}Nov 15, 2020'),
    (
        '2020-09-01/2020-11-15',
        'en',
        'full',
        f'Tue, Sep 1{separator}Sun, Nov 15, 2020'
    ),
    # ### Intervals (different precision)
    ('2020-09-02/2020-11', 'en', 'long', f'Sep 2{separator}Nov 30, 2020'),
    ('2020-09/2020-11-15', 'en', 'long', f'Sep 1{separator}Nov 15, 2020'),
])
def test_format_edtf(edtfstr, locale, format, expected):
    assert format_edtf(edtfstr, format=format, locale=locale) == expected


def test_format_edtf_invalid():
    """Test invalid values to format_edtf."""
    pytest.raises(ValueError, format_edtf, 'invalid')

    # With version 5.0.0, this is parsed as an interval,
    # but `interval.lower` is of type `UncertainOrApproximate`,
    # so the precision cannot be determined for formatting,
    # hence the `AttributeError`.
    pytest.raises(AttributeError, format_edtf, '2021/')

    pytest.raises(ValueError, format_edtf, '2020?')
    pytest.raises(TypeError, format_edtf, 2020)


def test_edtf_to_datetime():
    """Test datetime conversion."""
    y = parse_edtf_level0('2020')
    assert edtf_to_datetime(y, 'lower') == datetime(2020, 1, 1)
    assert edtf_to_datetime(y, 'upper') == datetime(2020, 12, 31)
    pytest.raises(ValueError, edtf_to_datetime, y, 'invalid')


def test_edtf_to_datetime_limit():
    """Test lowest possible date."""
    y = parse_edtf_level0('1000')
    assert edtf_to_datetime(y, 'lower') == datetime(1000, 1, 1)
    assert edtf_to_datetime(y, 'upper') == datetime(1000, 12, 31)


def test_format_edtf_default():
    """Test the default time being used if not value provided."""
    today = datetime.utcnow().date()
    assert format_edtf(format='short', locale='en') == '{}/{}/{}'.format(
        today.month, today.day, str(today.year)[2:4]
    )


def test_format_edtf_format():
    """Test overriding the format."""
    assert format_edtf('2020-11', format='yMd', locale='en') == '11/1/2020'
    assert format_edtf('2020/2021', format='yM', locale='en') == \
        f'1/2020{separator}12/2021'


@pytest.mark.parametrize('edtfstr,expected', [
    # ## Dates
    # Year in range
    ('2020', Date('2020')),
    ('2020/2021', Interval(Date('2020'), Date('2021'))),
    # Month in range
    ('2020-12', Date('2020', '12')),
    ('2020-09/2020-12', Interval(Date('2020', '09'), Date('2020', '12'))),
    # Day in range
    ('2020-12-31', Date('2020', '12', '31')),
    (
        '2020-09-01/2020-12-31',
        Interval(Date('2020', '09', '01'), Date('2020', '12', '31'))
    ),
    # Day in range for month of February on leap year
    ('2020-02-29', Date('2020', '02', '29')),
    (
        '2020-02-29/2020-12-31',
        Interval(Date('2020', '02', '29'), Date('2020', '12', '31'))
    ),
    # ## Dates and times
    # Complete representations for calendar date and (local) time of day
    (
        '1985-04-12T23:20:30',
        DateAndTime(Date('1985', '04', '12'), '23:20:30')
    ),
    # Complete representations for calendar date and UTC time of day
    (
        '1985-04-12T23:20:30Z',
        DateAndTime(Date('1985', '04', '12'), '23:20:30Z')
    ),
    # Date and time with timeshift in hours (only)
    (
        '1985-04-12T23:20:30-04',
        DateAndTime(Date('1985', '04', '12'), '23:20:30-04')
    ),
    # Date and time with timeshift in hours and minutes
    (
        '1985-04-12T23:20:30+04:30',
        DateAndTime(Date('1985', '04', '12'), '23:20:30+04:30')
    ),
])
def test_parse_edtf(edtfstr, expected):
    """Test valid values to parse_edtf."""
    assert parse_edtf(edtfstr) == expected


@pytest.mark.parametrize('edtfstr', [
    # Month out of range
    ('2020-13'),
    ('2020-13/2021-11'),
    ('2020-09/2021-13'),
    ('2020-13-01'),
    ('2020-09-01/2020-13-01'),
    ('2020-13-01/2020-09-01'),
    # Day out of range for month
    ('2020-04-31'),
    ('2020-04-31/2020-11-15'),
    ('2020-09-01/2020-04-31'),
    ('2020-02-30'),
    ('2020-02-30/2020-11-15'),
    ('2020-09-01/2020-02-30'),
    # Day out of range for month of February on non-leap year
    ('2021-02-29'),
    ('2021-02-29/2020-11-15'),
    ('2020-09-01/2021-02-29'),
])
def test_parse_edtf_invalid(edtfstr):
    """Test invalid values to parse_edtf."""
    pytest.raises(EDTFParseException, parse_edtf, edtfstr)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020-2024 CERN.
#
# Babel-EDTF is free software; you can redistribute it and/or modify it under
# the terms of the MIT License; see LICENSE file for more details.

"""Version information for Babel-EDTF.

This file is imported by ``babel_edtf.__init__``,
and parsed by ``setup.py``.
"""

__version__ = '1.2.0'



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020-2024 CERN.
#
# Babel-EDTF is free software; you can redistribute it and/or modify it under
# the terms of the MIT License; see LICENSE file for more details.

"""Localization of Extended Date Time Format (EDTF) level 0 strings."""

import calendar
from datetime import date as date_

from babel import Locale
from babel.dates import LC_TIME, format_date, format_interval, format_skeleton
from dateutil.parser import isoparse
from edtf import Date, EDTFObject, Interval
from edtf import parse_edtf as edtf_parse_edtf
from edtf import struct_time_to_datetime
from edtf.parser.grammar import EDTFParseException, ParseException
from edtf.parser.parser_classes import PRECISION_DAY, PRECISION_MONTH, \
    PRECISION_YEAR

from .version import __version__

BOUND_LOWER = 'lower'
BOUND_UPPER = 'upper'


DATE_SKELETON_FORMATS = {
    PRECISION_YEAR: {
        'full': 'y',
        'long': 'y',
        'medium': 'y',
        'short': 'y',
    },
    PRECISION_MONTH: {
        'full': 'yMMMM',
        'long': 'yMMMM',
        'medium': 'yMMM',
        'short': 'yM',
    },
    # Day precision is only used for intervals (not for date or date and time)
    # formatting. This is because the format_skeleton does not format exactly
    # the same way as format_date, and thus we use format_date to ensure all
    # dates look the same independently if they where formatted with
    # format_edtf or format_date.
    PRECISION_DAY: {
        'full': 'EEEEyMMMMd',
        'long': 'yMMMMd',
        'medium': 'yMMMd',
        'short': 'yMd',
    },
}


class EDTFValueError(ValueError):
    """An error for invalid EDTF formatted strings."""


class EDTFTypeError(TypeError):
    """An error for invalid EDTF formatted strings."""


def parse_edtf_level0(edtfstr):
    """Parse EDTF input string."""
    try:
        return parse_edtf(edtfstr)
    except ParseException:
        raise EDTFValueError(
            "The string is not a valid EDTF-formatted string.")


def get_edtf_date_skeleton(precision, format='medium'):
    """Return the date skeleton for a given precision.

    :param precision: the precision to use, one of "year", "month" or "day"
    :param format: the format to use, one of "full", "long", "medium", or
                   "short"
    """
    return DATE_SKELETON_FORMATS[precision][format]


def get_interval_precision(interval):
    """Get the precision for an interval."""
    precisions = [interval.lower.precision, interval.upper.precision]
    for p in [PRECISION_DAY, PRECISION_MONTH, PRECISION_YEAR]:
        if p in precisions:
            return p


def edtf_to_datetime(edtf_date, strict):
    """Convert an EDTF date to a Python date object."""
    if strict == BOUND_LOWER:
        date = edtf_date.lower_strict()
    elif strict == BOUND_UPPER:
        date = edtf_date.upper_strict()
    else:
        raise ValueError("Invalid value for 'strict' parameter.")

    return struct_time_to_datetime(date)


def format_edtf(edtf_level0=None, format='medium', locale=LC_TIME):
    """Format a EDTF level 0 expression.

    The formatting relies on Babel's ``format_skeleton()`` and
    ``format_interval()`` for all the heavy lifting.

    :param edtf_level0: a Date, Interval, or string representing
        and EDTF level 0 expression.
    :param format: one of "full", "long", "medium", or "short", or a custom
                   date/time pattern
    :param locale: a `Locale` object or a locale identifier
    """
    if edtf_level0 is None:
        edtf_level0 = date_.today().isoformat()

    if isinstance(edtf_level0, str):
        edtf_level0 = parse_edtf_level0(edtf_level0)

    # Do we have an EDTFObject (directly or parsed from a string)?
    if not isinstance(edtf_level0, EDTFObject):
        raise EDTFTypeError(
            'You must provide either a EDTF formatted string or an EDTF '
            'object.')

    # Do we have a Date, Interval or DateAndTime?
    if not (isinstance(edtf_level0, Date)
            or isinstance(edtf_level0, Interval)):
        raise EDTFValueError(
            'Only an EDTF level 0 date, date and time or interval is '
            'supported.'
        )

    locale = Locale.parse(locale)

    if isinstance(edtf_level0, Date):
        return _format_edtf0_date(edtf_level0, format, locale, BOUND_LOWER)
    elif isinstance(edtf_level0, Interval):
        return _format_edtf0_interval_naive(edtf_level0, format, locale)


def _format_edtf0_date(edtf_date, format, locale, strict):
    """Format an EDTF level 0 date."""
    # Convert EDTFDate to a python Date
    dt = edtf_to_datetime(edtf_date, strict)

    if edtf_date.precision == PRECISION_DAY:
        # Day precision: use normal date formatter
        return format_date(dt.date(), format=format, locale=locale)
    else:
        # Year or month precision: Use skeleton formatting
        if format in ('full', 'long', 'medium', 'short'):
            skeleton = get_edtf_date_skeleton(edtf_date.precision, format)
        else:
            skeleton = format

        return format_skeleton(
            skeleton, datetime=dt, fuzzy=True, locale=locale)


def _format_edtf0_interval_naive(edtf_interval, format, locale):
    """Format an EDTF level 0 interval."""
    dt_start = edtf_to_datetime(edtf_interval.lower, BOUND_LOWER).date()
    dt_end = edtf_to_datetime(edtf_interval.upper, BOUND_UPPER).date()

    precision = get_interval_precision(edtf_interval)

    # Get skeleton format
    if format in ('full', 'long', 'medium', 'short'):
        skeleton = get_edtf_date_skeleton(precision, format)
    else:
        skeleton = format

    return format_interval(
        dt_start, dt_end, skeleton, fuzzy=True, locale=locale)


def _validate_leap_year(edtf_date):
    """Validate that dates on the 29th of February are on leap years."""
    if isinstance(edtf_date, Date):
        if (
            edtf_date.precision == PRECISION_DAY
            and edtf_date.day == '29'
            and edtf_date.month == '02'
            and not calendar.isleap(int(edtf_date.year))
        ):
            raise EDTFParseException(
                'Day is out of range for month of February on non-leap year.'
            )
    elif isinstance(edtf_date, Interval):
        _validate_leap_year(edtf_date.lower)
        _validate_leap_year(edtf_date.upper)


def parse_edtf(date):
    """parse_edtf after trying isoparse for simple date formats.

    parse_edtf/pyparsing don't work in a thread safe way and throw
    TypeError randomly on some runs. This function tries to get isoparse first
    and then falls back to parse_edtf
    """
    if len(date) == 10:
        try:
            isodate = isoparse(date)
            return Date(
                str(isodate.year),
                str(isodate.month).zfill(2),
                str(isodate.day).zfill(2),
            )
        except Exception:
            pass

    edtf_date = edtf_parse_edtf(date)

    # python-edtf's edtf_parse_edtf uses a grammar that allows days up to 29
    # for the month of February regardless of whether the year is a leap year
    # or not. Here we fail in case we hit this corner case.
    _validate_leap_year(edtf_date)

    return edtf_date
